# 0.3.0 developer notes
Version 0.3.0 of NearestNeighborDescent.jl will consist of a major overhaul
of the structures used in the nndescent algorithm. There are two high level
goals:

1. Add support for shared memory (multi-threaded) parallelism, relying on the
new parallel task runtime introduced in Julia 1.3
2. Add a collection of KNN graph types that implement the `AbstractGraph`
interface

There are also several sub-goals / requirements:
- At least one of the KNN graph implementations must be thread-safe
- If the KNN graphs themselves support parallelism (decision to be made later),
this should be designed to require as little coordination with the parallelism
of the algorithm as possible (parallelism designed to be nested)

## Parallel Nearest Neighbor Descent
For data set V, integer k, and distance function σ:

```julia
function NNDescent(V, k, σ)
  G = ApproximateKNNGraph(V, k, σ)  # randomly initialize G
  while true
    N = [neighbors(G, v) for v in vertices(G)] # forward and backward
    c = 0
    parallel for v in vertices(G)
      parallel for p in N[v], q in N[v] # perform a local join around each point
        w = σ(p, q)
        c = c + add_edge!(G, p, q, w)
      end
    end
    if c == 0
      return G
    end
  end
```

### Potential Functions
*(names not final)*
- `refine` and `refine!`: Provide an interface for performing a single iteration
on a knn graph, one that that mutates (previous behavior) and one that doesn't.
The non-mutating `refine` could be used to inspect the graph after each
refining step; to plot, or to check if candidates are improving, etc...

## Approximate *k*-Nearest Neighbor Graphs
**Definition**: An approximate k-nearest neighbor graph is a weighted, directed
graph *G = {V, E, k}*, where *V* is a non-empty set of nodes, *E* is a set of
weighted directed edges *(p, q, w)*, and *k* is a positive integer, with the
following properties:
1. | *V* | >  *k*
2. Every node *v* in *V* has exactly *k* outgoing edges, with no self-loops.
  - This implies that | *E* | = | *V* | * *k*.
The edges *(p, q)* in *E* denote that *q* is an *approximate* (or candidate)
nearest neighbor of *p* in *V*.

For the types in this package, the set of nodes is immutable but the set of
outgoing edges for each node can be updated such that the sum of all outgoing
edge weights never increases.

### Types
- **`ApproximateKNNGraph <: AbstractGraph`**: An abstract type whose subtypes
all have the properties of approximate knn graphs
- **`HeapKNNGraph <: ApproximateKNNGraph`**: A knn graph where each node's
outgoing edges are stored in a max heap
  - NOTE: This isn't more generic (e.g. `HeapKNNGraph{h <: AbstractMaxHeap}`)
  specifically because `DataStuctures.jl` needs a refactoring to using the
  `Base.Ordering` types instead of its own `LessThan`/`GreaterThan` ordering
  types. I want to avoid relying on any outdated abstractions from that package
  that really ought to be refactored.
- **`ThreadSafeHeapKNNGraph`**: A thread-safe version of `HeapKNNGraph`
(**NOTE**: use traits?)

#### Internal Types
(e.g. edge types, node types, required by `LightGraphs`).

### Functionality
- `ApproximateKNNGraph(D, k, σ)`: A constructor that takes the set of points, k,
and initializes the knn graph by randomly selecting k points and calculating
weights with σ.
- `ApproximateKNNGraph(initial)`: A constructor where the `initial` k
candidate neighbors of each point are provided by the caller.
- `knn_diameter(g, v; dir=:out)`: Return the diameter of the ball centered at
vertex `v` containing all the neighbors of `v`. `dir` can be used to specify
which neighbors to consider, one of {`:in`, `:out`, `:both`}.

## Public API and Usage
Anything that is `export`ed should be considered part of the public API for this
package and subject to [semver](semver.org) versioning. 

### Usage #1 - `nndescent`
The likely most common way this package will be used is by one-shot approximate
kNN graph construction on a dataset:
```julia
nndescent(data, n_neighbors, metric; max_iters, sample_rate, precision) -> KNNGraph
```
Since this is an `AbstractGraph`, all the usual `LightGraphs` utilities will work 
on it. Convenience functions for KNN Graphs are also exported:
```julia
# return the approximate knns as matrices of indexes and distances, where
# indices[j, i] and distances[j, i] are the index of and distance to node i's jth
# nearest neighbor, respectively.
knn_matrices(graph) -> indices, distances

knn_diameter(graph, v) -> diameter
```

### Usage #2 - `search`
Another common use pattern will be finding the approx KNNs in an already constructed
graph. 
```julia
search(graph, data, queries, n_neighbors) -> indices, distances
```

